/*
 * Created on 12.05.2006 for PIROL
 *
 * SVN header information:
 *  $Author: LBST-PF-3\orahn $
 *  $Rev: 2516 $
 *  $Date: 2006-10-09 16:56:06 +0200 (Mo, 09 Okt 2006) $
 *  $Id: DelaunayDiagrammTools.java 2516 2006-10-09 14:56:06Z LBST-PF-3\orahn $
 */
package de.fhOsnabrueck.jump.pirol.utilities.graph.Delaunay;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;

import de.fhOsnabrueck.jump.pirol.utilities.apiTools.FeatureCollectionTools;
import de.fhOsnabrueck.jump.pirol.utilities.comparisonAndSorting.ObjectComparator;
import de.fhOsnabrueck.jump.pirol.utilities.debugOutput.DebugUserIds;
import de.fhOsnabrueck.jump.pirol.utilities.debugOutput.PersonalLogger;
import de.fhOsnabrueck.jump.pirol.utilities.graph.DreiecksInterpolierer;
import de.fhOsnabrueck.jump.pirol.utilities.graph.PirolEdge;
import de.fhOsnabrueck.jump.pirol.utilities.graph.PirolPoint;

/**
 * class containing methods to work on an delaunay diagramm,
 * like creating isolines or gradients.
 *
 * @author Ole Rahn
 * <br>
 * <br>FH Osnabr&uuml;ck - University of Applied Sciences Osnabr&uuml;ck,
 * <br>Project: PIROL (2006),
 * <br>Subproject: Daten- und Wissensmanagement
 * 
 * @version $Rev: 2516 $
 * 
 */
public abstract class DelaunayDiagrammTools {
    
    protected FeatureCollection featureCollection = null;
    protected HashMap fid2Feature = new HashMap();
    protected DelaunayPunkt[] pointsArray = null;
    
    protected int indexOfAttributeOfInterest;
    
    protected PersonalLogger logger = new PersonalLogger(DebugUserIds.OLE);
    
    /**
     * 
     *@param featureCollection the feature collection of the layer the Delaunay diagram was calculated for
     *@param pointsArray array of Delaunay points
     *@param indexOfAttributeOfInterest index of the attribute to work with
     */
    public DelaunayDiagrammTools(FeatureCollection featureCollection, DelaunayPunkt[] pointsArray, int indexOfAttributeOfInterest ) {
        this.featureCollection = featureCollection;
        this.pointsArray = pointsArray;
        this.indexOfAttributeOfInterest = indexOfAttributeOfInterest;
    }
    
    protected List findPointWithValueOnLine(PirolEdge line, double isoLiniesValue){
        List result = new ArrayList();
        
        Feature anfangsPunkt = this.getFeature( line.getAnfang().getIndex() );
        Feature endPunkt = this.getFeature( line.getEnde().getIndex() );
        
        double wertAnfang = ObjectComparator.getDoubleValue(anfangsPunkt.getAttribute(this.indexOfAttributeOfInterest));
        double wertEnde = ObjectComparator.getDoubleValue(endPunkt.getAttribute(this.indexOfAttributeOfInterest));
        
        if (wertAnfang == Double.NaN || wertEnde == Double.NaN){
            logger.printDebug( this.getClass().getName() + ".findeInterpolationsPunkt: found NAN!");
        }
        
        double grosserWert = Math.max(wertAnfang, wertEnde);
        double kleinerWert = Math.min(wertAnfang, wertEnde);
        
        if ( grosserWert < isoLiniesValue || kleinerWert > isoLiniesValue ) {
            return null;            
        } else if ( grosserWert == kleinerWert ) {
            // must be == this.isoLinienWert
            result.add(line.getAnfang());
            result.add(line.getEnde());
        } else if ( wertAnfang == isoLiniesValue ) {
            result.add(line.getAnfang());
        } else if ( wertEnde == isoLiniesValue ) {
            result.add(line.getEnde());
        } else {
            PirolPoint a,b;
            if (wertAnfang < isoLiniesValue){
                a = line.getAnfang();
                b = line.getEnde();
            } else {
                a = line.getEnde();
                b = line.getAnfang();
            }
            try {
                PirolPoint ab = new PirolPoint( new double[]{b.getX()-a.getX(), b.getY()-a.getY()}, -1);
                double abFaktor = this.getLengthFactorForValueOnLine( kleinerWert, grosserWert, isoLiniesValue );
                PirolPoint iPunkt = new PirolPoint( new double[]{a.getX()+abFaktor*ab.getX(), a.getY()+abFaktor*ab.getY()}, -1 );
                
                result.add(iPunkt);
            } catch (Exception e ){
                e.printStackTrace();                
            }
        }
        return result;
    } 
    
    protected double getLengthFactorForValueOnLine( double greatValue, double smallValue, double value ){
        double wertDifferenz = smallValue - greatValue;
        double faktorGrundlage = value - greatValue;
        return (faktorGrundlage/wertDifferenz);
    }
    
    protected double interpolateValueOnLine( PirolEdge line, double distanceFromStartingPointOnLine ) throws Exception{
        Feature startingPoint = FeatureCollectionTools.getFeatureFromCollection(this.featureCollection, line.getAnfang().getIndex());
        Feature endPoint = FeatureCollectionTools.getFeatureFromCollection(this.featureCollection, line.getEnde().getIndex());
        
        double lengthOfLine = line.getLaenge();
        double valueA = ((Double)startingPoint.getAttribute(this.indexOfAttributeOfInterest)).doubleValue();
        double valueB = ((Double)endPoint.getAttribute(this.indexOfAttributeOfInterest)).doubleValue();
        
        double factor = distanceFromStartingPointOnLine/lengthOfLine;
        
        if (factor<0 || factor>1){
            logger.printError("invalid factor");
            return Double.NaN;
        }
        
        double diff = valueB - valueA;
        
        return valueA + (diff * factor);        
    }
    
    protected List searchForFurtherLines( DelaunayPunkt punkt1, DelaunayPunkt punkt2 ){
        
        ArrayList suspekteKanten = new ArrayList();
        
        suspekteKanten.addAll(punkt1.getBildetKanten());
        suspekteKanten.addAll(punkt2.getBildetKanten());
        
        ArrayList passendeKanten = new ArrayList();
        
        PirolEdge ki, kj;
        
        for ( int i=0; i<suspekteKanten.size(); i++ ){
            ki = (PirolEdge)suspekteKanten.get(i);
            for ( int j=0; j<suspekteKanten.size(); j++ ){
                if (i == j) continue;
                kj = (PirolEdge)suspekteKanten.get(j);
                
                if ( 
                    ( kj.getAnfang().equals( ki.getAnfang()) && !kj.getAnfang().equals( punkt1 ) && !kj.getAnfang().equals( punkt2 ) ) ||
                    ( kj.getEnde().equals(   ki.getEnde()  ) && !kj.getEnde().equals(   punkt1 ) && !kj.getEnde().equals(   punkt2 )) ||
                    ( kj.getEnde().equals(   ki.getAnfang()) && !kj.getEnde().equals(   punkt1 ) && !kj.getEnde().equals(   punkt2 )) ||
                    ( kj.getAnfang().equals( ki.getEnde()  ) && !kj.getAnfang().equals( punkt1 ) && !kj.getAnfang().equals( punkt2 ))
                    ){
                    if ( !passendeKanten.contains(ki) )
                        passendeKanten.add(ki);
                    if ( !passendeKanten.contains(kj) )
                        passendeKanten.add(kj);
                }
            }
        }
        
        return passendeKanten;
    }

    protected double interpolateValueWithinTriangle( PirolPoint testPunkt, PirolEdge kante1, PirolEdge kante2, PirolEdge kante3 ) throws Exception{
        //  ----->       -------->      -------->   -------->
        //  punkt1 + a * punkt_1_2 + b* punkt_2_3 = testPunkt
        //  a und b fuer x und y loesen, dann 
        
        PirolPoint punkt1, punkt2, punkt3;
        PirolPoint[] punkte = DelaunayDiagrammTools.getArrangedPunktArray(new PirolEdge[]{kante1,kante2,kante3});
        
        punkt1 = punkte[0];
        punkt2 = punkte[1];
        punkt3 = punkte[2];
        
        double a = 0.;
        double b = 0.;
        
        PirolPoint punkt_1_2 = new PirolPoint( new double[]{punkt2.getX() - punkt1.getX(), punkt2.getY() - punkt1.getY()}, -1 );
        PirolPoint punkt_2_3 = new PirolPoint( new double[]{punkt3.getX() - punkt2.getX(), punkt3.getY() - punkt2.getY()}, -1 );
        
        if ( punkt_1_2.getX() != 0 && punkt_2_3.getY() != (punkt_2_3.getX()*punkt_1_2.getY()/punkt_1_2.getX()) )
            b = (punkt1.getY()-testPunkt.getY()-((punkt1.getX()-testPunkt.getX())*punkt_1_2.getY()/punkt_1_2.getX()))/(-punkt_2_3.getY()+(punkt_2_3.getX()*punkt_1_2.getY()/punkt_1_2.getX()));
        else if ( punkt_2_3.getY() != 0 )
            b = ( punkt1.getX() - testPunkt.getX() )/(-punkt_2_3.getX());
        
        if ( punkt_1_2.getX() != 0 )
            a = (punkt1.getX()+b*punkt_2_3.getX()-testPunkt.getX())/(-punkt_1_2.getX());
        else if ( punkt_1_2.getY() != 0 && (-punkt_2_3.getX())*punkt_2_3.getY() != testPunkt.getY() )
            a = (punkt1.getY()+(punkt1.getX()-testPunkt.getX())/(-punkt_2_3.getX())*punkt_2_3.getY()-testPunkt.getY() )/(-punkt_1_2.getY());
            
        // hole werte:

        Feature feat1 = FeatureCollectionTools.getFeatureFromCollection(this.featureCollection, punkt1.getIndex());
        Feature feat2 = FeatureCollectionTools.getFeatureFromCollection(this.featureCollection, punkt2.getIndex());
        Feature feat3 = FeatureCollectionTools.getFeatureFromCollection(this.featureCollection, punkt3.getIndex());
        
        double wert1 = ObjectComparator.getDoubleValue(feat1.getAttribute(this.indexOfAttributeOfInterest));
        double wert2 = ObjectComparator.getDoubleValue(feat2.getAttribute(this.indexOfAttributeOfInterest));
        double wert_1_2 = wert2 - wert1;
        double wert3 = ObjectComparator.getDoubleValue(feat3.getAttribute(this.indexOfAttributeOfInterest));
        double wert_2_3 = wert3 - wert2;
        
        if (wert1 == Double.NaN || wert2 == Double.NaN || wert3 == Double.NaN){
            logger.printDebug( this.getClass().getName() + ".interpoliereWertInDreieck: found NAN!");
        }
        
        return ( wert1 + a*wert_1_2 + b*wert_2_3 );
    }

    protected void findLinesBeginningAtLine( PirolEdge testKante, List neueKanten, double value ){

        DelaunayPunkt punkt1 = (DelaunayPunkt)testKante.getAnfang();
        DelaunayPunkt punkt2 = (DelaunayPunkt)testKante.getEnde();
        
        double wertAnfang = ObjectComparator.getDoubleValue(FeatureCollectionTools.getFeatureFromCollection(this.featureCollection, punkt1.getIndex()).getAttribute(this.indexOfAttributeOfInterest));
        double wertEnde   = ObjectComparator.getDoubleValue(FeatureCollectionTools.getFeatureFromCollection(this.featureCollection, punkt2.getIndex()).getAttribute(this.indexOfAttributeOfInterest));
        
        if (Double.isNaN(wertAnfang) || Double.isNaN(wertEnde)){
            logger.printWarning( this.getClass().getName() + ".findeLinienBeginnendAnKante: found NAN!");
        }
        
        PirolPoint neueLinie_anfangspunkt = null;

        List moeglichePunkte = this.findPointWithValueOnLine( testKante, value );
        
        if ( moeglichePunkte.size() == 2 ){
            neueLinie_anfangspunkt = (DelaunayPunkt)moeglichePunkte.get(0);
        } else if ( moeglichePunkte.size() == 1 ){
            neueLinie_anfangspunkt = (PirolPoint)moeglichePunkte.get(0);
            List suspekteKanten = this.searchForFurtherLines( punkt1, punkt2 );

            List moeglicheEndpunkte = new ArrayList();
            Iterator iter = suspekteKanten.iterator();
            PirolEdge k;
            DelaunayPunkt p1, p2;
            
            while ( iter.hasNext() ){
                k = (PirolEdge)iter.next();
                List moeglicheEndpunkteTmp = this.findPointWithValueOnLine( k, value );
                
                if (moeglicheEndpunkteTmp==null) continue;
                
                if ( moeglicheEndpunkteTmp.size() ==2 ){
                    p1 = (DelaunayPunkt)moeglicheEndpunkteTmp.get(0);
                    p2 = (DelaunayPunkt)moeglicheEndpunkteTmp.get(1);
                    PirolEdge neueKante = new PirolEdge( p1, p2 );
                    neueKanten.add( neueKante );
                } else {
                    // kann im Grunde nur 1 Punkt sein...
                    for (int j=0; j<moeglicheEndpunkteTmp.size(); j++){
                        moeglicheEndpunkte.add(moeglicheEndpunkteTmp.get(j));
                    }
                }
            }
            if ( moeglicheEndpunkte.isEmpty() ){
                return;
            
            } else if ( moeglicheEndpunkte.size() >= 1 ){

                for (int i=0; i<moeglicheEndpunkte.size(); i++ ){
                    PirolEdge neueKante = new PirolEdge( neueLinie_anfangspunkt, (PirolPoint)moeglicheEndpunkte.get(i) );
                    neueKanten.add( neueKante );
                }
            }
        }
    }
    
    protected PirolEdge findConnectingLine( PirolEdge k1, PirolEdge k2 ){
        DelaunayPunkt p1=null, p2=null;
        
        if ( k1.getAnfang().equals(k2.getAnfang())){
            p1 = (DelaunayPunkt)k1.getEnde();
            p2 = (DelaunayPunkt)k2.getEnde();
        } else if ( k1.getAnfang().equals(k2.getEnde())){
            p1 = (DelaunayPunkt)k1.getEnde();
            p2 = (DelaunayPunkt)k2.getAnfang();
        } else if ( k1.getEnde().equals(k2.getAnfang())){
            p1 = (DelaunayPunkt)k1.getAnfang();
            p2 = (DelaunayPunkt)k2.getEnde();
        } else if ( k1.getEnde().equals(k2.getEnde())){
            p1 = (DelaunayPunkt)k1.getAnfang();
            p2 = (DelaunayPunkt)k2.getAnfang();
        } else 
            return null;
        
        PirolEdge knte;
        if (p1.isNeighborOf(p2)){
            PirolEdge[] kanten = p1.getBildetKantenArray();
            for (int i=0; i<kanten.length; i++){
                knte = kanten[i];
                if ( (knte.getAnfang().equals(p1) || knte.getAnfang().equals(p2)) && (knte.getEnde().equals(p1) || knte.getEnde().equals(p2))){
                    return knte;
                }
            }
        }
            
        
        return null;
    }
    
    public final static boolean isPointWithinTriangle( PirolPoint testPunkt, PirolEdge kante1, PirolEdge kante2, PirolEdge kante3) throws Exception{

        PirolPoint[] punkte = DelaunayDiagrammTools.getArrangedPunktArray(new PirolEdge[]{kante1, kante2, kante3});
        
        if (punkte.length != 4)
            throw new Exception("wrong point count");
        
        PirolEdge k1 = new PirolEdge( punkte[0], punkte[1] );
        PirolEdge k2 = new PirolEdge( punkte[1], punkte[2] );
        PirolEdge k3 = new PirolEdge( punkte[2], punkte[3] );
        
        double faktor1 = k1.getNormalenFaktorZu(testPunkt);
        double faktor2 = k2.getNormalenFaktorZu(testPunkt);
        double faktor3 = k3.getNormalenFaktorZu(testPunkt);
        
        boolean allLessThanOrEqualToZero = (faktor1<=0) && (faktor2<=0) && (faktor3<=0); 
        boolean allGreaterThanZero = (faktor1>0) && (faktor2>0) && (faktor3>0);
       
        return allLessThanOrEqualToZero || allGreaterThanZero;
    }
    
    protected final static PirolPoint[] getArrangedPunktArray( PirolEdge[] kanten ){
        PirolPoint[] punkte = new PirolPoint[kanten.length+1];
        punkte[0] = kanten[0].getAnfang();
        punkte[1] = kanten[0].getEnde();
        kanten[0].setGueltig(false);
        
        PirolPoint letzterPunkt = kanten[0].getEnde();
        PirolPoint tmp1, tmp2;
        
        int punkteIndex = 1;
        
        for (int j=1; j<kanten.length && !punkte[0].liegtAuf(letzterPunkt); j++){
            for ( int i=1; i<kanten.length; i++){
                if ( kanten[i]==null || !kanten[i].isGueltig()) continue;
                tmp1 = kanten[i].getAnfang();
                tmp2 = kanten[i].getEnde();
                
                if (tmp1.liegtAuf(letzterPunkt)){
                    punkte[++punkteIndex] = tmp2;
                    letzterPunkt = tmp2;
                    kanten[i].setGueltig(false);
                } else if (tmp2.liegtAuf(letzterPunkt)){
                    punkte[++punkteIndex] = tmp1;
                    letzterPunkt = tmp1;
                    kanten[i].setGueltig(false);
                }
            }
        }
        for ( int i=0; i<kanten.length; i++){
            kanten[i].setGueltig(true);
        }
        
        return punkte;
    }
    
    public final static LinearRing kanten2LinearRing( PirolEdge[] kanten ) throws Exception{
        PirolPoint[] punkte = DelaunayDiagrammTools.getArrangedPunktArray( kanten );
        Coordinate[] coords = new Coordinate[punkte.length];
        
        if (punkte == null)
            throw new Exception("no points array could be extracted out of the lines array");
        
        for ( int i=0; i<punkte.length; i++){
            if (punkte[i] == null)
                throw new Exception(DreiecksInterpolierer.class.getName() + ".kanten2LinearRing(): null point found: " + i);
            coords[i] = new Coordinate(punkte[i].getX(), punkte[i].getY());
        }
        
        GeometryFactory geoFactory = new GeometryFactory();
        
        return geoFactory.createLinearRing(coords);
        
    }
    
    protected Feature findPointNear( double x, double y ){
        Feature feat = null;
        
        double step = 0.05;
        double difference = 0.0;
        
        GeometryFactory gf = new GeometryFactory();
        Point p = gf.createPoint(new Coordinate(x,y));
        
        if ( x < 20 || y < 20 ){
            // WGS84 or so - at least not GK
            step = 0.0000005;
        }
        
        Feature[] result = null; 
        
        Envelope env;
        
        while (result == null || result.length == 0){
            env = new Envelope(x-difference,x+difference,y-difference,y+difference);
            try {
                result = (Feature[])this.featureCollection.query(env).toArray(new Feature[0]);
            } catch (RuntimeException e) {
                continue;
            }
            difference += step;
        }
        
        Feature thisFeature;
        double minDist = Double.MAX_VALUE, aktDist;
        
        for (int i=0; i<result.length; i++){
            thisFeature = result[i];
            aktDist = thisFeature.getGeometry().distance(p);
            
            if (aktDist < minDist){
                feat = thisFeature;
                minDist = aktDist;
            }
            
        }
        
        return feat;
    }

    protected double calculateValueAtPoint( PirolPoint ausgangsPunkt, DelaunayPunkt[] punktArray ) throws Exception{
    
        Feature naechsterPunkt = this.findPointNear( ausgangsPunkt.getX(), ausgangsPunkt.getY() );
        int npIndex = naechsterPunkt.getID();
        
        double value = Double.NaN;

        if (naechsterPunkt == null){
            return Double.NaN;
        }
        
        logger.printDebug("berechneWertAnPunkt " + ausgangsPunkt.toString() + " -> " + naechsterPunkt.getID());
        
        DelaunayPunkt nP = null;
        
        DelaunayPunkt tmp;
        for (int i=0; i<punktArray.length && nP==null; i++){
            tmp = punktArray[i];
            if (tmp.getIndex() == npIndex){
                nP = tmp;
                break;
            }
            
        }
        
        if (nP == null){
            throw new Exception("can not find next point feature to calculate attribute value at the position of mouse cursor");
        }
//        ArrayList evtlBeteiligteKanten = nP.getBildetKanten();
        PirolEdge[] evtlBeteiligteKanten = (PirolEdge[])nP.getBildetKanten().toArray(new PirolEdge[0]);
        
        PirolEdge kante1 = null;
        PirolEdge kante2 = null;
        PirolEdge kante3 = null;
        
        PirolEdge ki, kj;
        
        for (int i=0; i<evtlBeteiligteKanten.length; i++ ){
            ki = evtlBeteiligteKanten[i];
            for (int j=0; j<evtlBeteiligteKanten.length; j++ ){
                if (i == j) continue;
                kj = evtlBeteiligteKanten[j];
                kante1 = null;
                kante2 = null;
                kante3 = null;
                
                kante3 = this.findConnectingLine( ki, kj );
                if (kante3 != null){
                    if ( DelaunayDiagrammTools.isPointWithinTriangle( ausgangsPunkt, ki, kj, kante3 ) ){
                        /*
                        logger.printDebug("liegtPunktImDreieck == true");
                        logger.printDebug(ki.toString());
                        logger.printDebug(kj.toString());
                        logger.printDebug(kante3.toString());
                        */
                        kante1 = ki;
                        kante2 = kj;
                        break;
                    }
                }
            }
                    
            if ( kante1 != null && kante2 != null )
                break;
        }
        if ( kante1 == null || kante2 == null || kante3 == null ){
            logger.printWarning( "- berechneWertAnPunkt(): konnte kein dreieck konstruieren - uebernehme wert des naechsten punktes:" );
            double wertAmNaechstenPunkt = ObjectComparator.getDoubleValue(naechsterPunkt.getAttribute(this.indexOfAttributeOfInterest));
            value = wertAmNaechstenPunkt;
        } else {
            value = this.interpolateValueWithinTriangle( ausgangsPunkt, kante1, kante2, kante3 );
        }
        
//        logger.printDebug("berechneWertAnPunkt: wert= " + this.isoLinienWert);
        return value;
    }        
    
    public int getIndexOfAttributeOfInterest() {
        return indexOfAttributeOfInterest;
    }
    public void setIndexOfAttributeOfInterest(int isoLinienAttribut) {
        this.indexOfAttributeOfInterest = isoLinienAttribut;
    }

    protected PirolPoint getCentroid(PirolPoint[] points) throws Exception{
        double[] coordSums = new double[points[0].getDimension()];
        
        for (int j=0; j<coordSums.length; j++){
            coordSums[j] = 0;
        }
        
        for ( int i=0; i<points.length; i++ ){
            for (int j=0; j<coordSums.length; j++){
                coordSums[j] += points[i].getCoordinate(j);
            }
        }
        
        for (int j=0; j<coordSums.length; j++){
            coordSums[j] /= points.length;
        }
        
        return new PirolPoint(coordSums);
    }
    
    protected final Feature getFeature(int fid){
        Feature feat = null;
        Integer FID = new Integer(fid);
        if ( !fid2Feature.containsKey(FID) ){
            feat = FeatureCollectionTools.getFeatureFromCollection( this.featureCollection, fid );
            fid2Feature.put(FID, feat);
        } else {
            feat = (Feature)fid2Feature.get(FID);
        }
        return feat;
    }

}
